from functools import lru_cache

import docker
from docker.client import DockerClient
from docker.errors import DockerException

from ._timeouts import DOCKER_REQUEST_TIMEOUT_SECONDS
from .errors import DockerError


@lru_cache(maxsize=1)
def get_docker_client() -> DockerClient:
    try:
        return docker.from_env(timeout=DOCKER_REQUEST_TIMEOUT_SECONDS)
    except DockerException as exc:
        raise DockerError("Docker is not running or not reachable. Start Docker and retry.") from exc



from dataclasses import dataclass
from pathlib import Path

from aicage.cli_types import ParsedArgs
from aicage.config.agent.loader import load_agents
from aicage.config.base.loader import load_bases
from aicage.config.config_store import SettingsStore
from aicage.config.context import ConfigContext
from aicage.config.extensions.loader import load_extensions
from aicage.config.project_config import AgentConfig
from aicage.registry.image_selection.models import ImageSelection
from aicage.registry.image_selection.selection import select_agent_image
from aicage.runtime.mounts.resolver import resolve_mounts
from aicage.runtime.prompts.confirm import prompt_persist_docker_args
from aicage.runtime.run_args import MountSpec


@dataclass(frozen=True)
class RunConfig:
    project_path: Path
    agent: str
    context: ConfigContext
    selection: ImageSelection
    project_docker_args: str
    mounts: list[MountSpec]


def load_run_config(agent: str, parsed: ParsedArgs | None = None) -> RunConfig:
    store = SettingsStore()
    project_path = Path.cwd().resolve()
    # project_config_path = store.project_config_path(project_path)

    # with _lock_project_config(project_config_path):
    bases = load_bases()
    agents = load_agents(bases)
    project_cfg = store.load_project(project_path)
    context = ConfigContext(
        store=store,
        project_cfg=project_cfg,
        agents=agents,
        bases=bases,
        extensions=load_extensions(),
    )
    selection = select_agent_image(agent, context)
    agent_cfg = project_cfg.agents.setdefault(agent, AgentConfig())

    existing_project_docker_args: str = agent_cfg.docker_args
    if agent_cfg.base is None:
        agent_cfg.base = selection.base

    mounts = resolve_mounts(context, agent, parsed)

    _persist_docker_args(agent_cfg, parsed)
    store.save_project(project_path, project_cfg)

    return RunConfig(
        project_path=project_path,
        agent=agent,
        context=context,
        selection=selection,
        project_docker_args=existing_project_docker_args,
        mounts=mounts,
    )


def _persist_docker_args(agent_cfg: AgentConfig, parsed: ParsedArgs | None) -> None:
    if parsed is None or not parsed.docker_args:
        return
    existing = agent_cfg.docker_args
    if existing == parsed.docker_args:
        return

    if prompt_persist_docker_args(parsed.docker_args, existing):
        agent_cfg.docker_args = parsed.docker_args



from pathlib import Path

from aicage.config.agent._custom_loader import load_custom_agents
from aicage.config.agent._metadata import build_agent_metadata
from aicage.config.agent._validation import ensure_required_files
from aicage.config.agent.models import AgentMetadata
from aicage.config.base.models import BaseMetadata
from aicage.config.errors import ConfigError
from aicage.config.resources import find_packaged_path
from aicage.config.yaml_loader import load_yaml

_AGENT_DEFINITION_FILES: tuple[str, str] = ("agent.yaml", "agent.yml")


def load_agents(bases: dict[str, BaseMetadata]) -> dict[str, AgentMetadata]:
    builtin_agents = _load_builtin_agents(bases)
    custom_agents = load_custom_agents(bases)
    merged_agents = dict(builtin_agents)
    merged_agents.update(custom_agents)
    return merged_agents


def _load_builtin_agents(bases: dict[str, BaseMetadata]) -> dict[str, AgentMetadata]:
    agents_dir = _builtin_agents_dir()
    if not agents_dir.is_dir():
        raise ConfigError(f"Built-in agent directory '{agents_dir}' is missing.")

    agents: dict[str, AgentMetadata] = {}
    for entry in sorted(agents_dir.iterdir()):
        if not entry.is_dir():
            continue
        agent_name = entry.name
        agent_path = _find_agent_definition(entry)
        agent_mapping = load_yaml(agent_path)
        ensure_required_files(agent_name, entry)
        agents[agent_name] = build_agent_metadata(
            agent_name=agent_name,
            agent_mapping=agent_mapping,
            bases=bases,
            definition_dir=entry,
        )
    return agents


def _builtin_agents_dir() -> Path:
    dockerfile = find_packaged_path("agent-build/Dockerfile")
    return dockerfile.parent / "agents"


def _find_agent_definition(agent_dir: Path) -> Path:
    for filename in _AGENT_DEFINITION_FILES:
        candidate = agent_dir / filename
        if candidate.is_file():
            return candidate
    expected = ", ".join(_AGENT_DEFINITION_FILES)
    raise ConfigError(f"Agent '{agent_dir.name}' is missing {expected}.")



from pathlib import Path

from ._tty import ensure_tty_for_prompt


def prompt_for_missing_extensions(
    agent: str,
    missing: list[str],
    stored_image_ref: str,
    project_config_path: Path,
    other_projects: list[tuple[str, Path]],
) -> str:
    ensure_tty_for_prompt()
    print(f"[aicage] Missing extensions for '{agent}': {', '.join(sorted(missing))}.")
    if stored_image_ref:
        print(f"[aicage] Stored image ref: {stored_image_ref}")
    print(f"[aicage] Project config: {project_config_path}")
    if other_projects:
        print("[aicage] Other projects using this image:")
        for project_path, config_path in other_projects:
            print(f"  {project_path} -> {config_path}")
    return input("Choose 'exit' or 'fresh': ").strip().lower()



from aicage.docker.query import get_local_rootfs_layers


def base_layer_missing(base_image_ref: str, final_image_ref: str) -> bool | None:
    base_layers = get_local_rootfs_layers(base_image_ref)
    if base_layers is None:
        return None
    final_layers = get_local_rootfs_layers(final_image_ref)
    if final_layers is None:
        return None
    return base_layers[-1] not in final_layers



# supress missing import _version as that file is autogenerated by `hatch build --hooks-only`
# pyright: reportMissingImports=false
from importlib import metadata

try:
    from ._version import __version__
except ImportError:
    try:
        __version__ = metadata.version("aicage")
    except metadata.PackageNotFoundError:
        __version__ = "0.0.0"



from dataclasses import dataclass
from pathlib import Path

BUILD_LOCAL_KEY: str = "build_local"


@dataclass(frozen=True)
class BaseMetadata:
    from_image: str
    base_image_distro: str
    base_image_description: str
    build_local: bool
    local_definition_dir: Path



import subprocess
from pathlib import Path


def capture_stdout(command: list[str], cwd: Path | None = None) -> str | None:
    try:
        result = subprocess.run(
            command, check=True, capture_output=True, text=True, cwd=str(cwd) if cwd else None
        )
    except (FileNotFoundError, subprocess.CalledProcessError):
        return None
    return result.stdout



from datetime import datetime, timezone


def now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


def timestamp() -> str:
    return datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%SZ")



import json
import urllib.request
from collections.abc import Mapping
from typing import Any

from ._timeouts import REGISTRY_REQUEST_TIMEOUT_SECONDS
from .errors import RegistryDiscoveryError
from .types import RegistryApiConfig


def _fetch_pull_token_for_repository(api_config: RegistryApiConfig, repository: str) -> str:
    url = f"{api_config.registry_api_token_url}:{repository}:pull"
    data, _ = _fetch_json(url, None)
    token = data.get("token")
    if not token:
        raise RegistryDiscoveryError(f"Missing token while querying registry for {repository}.")
    return token


def _fetch_json(url: str, headers: dict[str, str] | None) -> tuple[dict[str, Any], Mapping[str, str]]:
    request = urllib.request.Request(url, headers=headers or {})
    try:
        with urllib.request.urlopen(request, timeout=REGISTRY_REQUEST_TIMEOUT_SECONDS) as response:
            payload = response.read().decode("utf-8")
            response_headers = response.headers
    except Exception as exc:  # pylint: disable=broad-except
        raise RegistryDiscoveryError(f"Failed to query registry endpoint {url}: {exc}") from exc

    try:
        data = json.loads(payload)
    except json.JSONDecodeError as exc:
        raise RegistryDiscoveryError(f"Invalid JSON from registry endpoint {url}: {exc}") from exc
    return data, response_headers



from dataclasses import dataclass

_DEFAULT_DOCKER_REGISTRY: str = "registry-1.docker.io"
_DOCKER_REGISTRY_ALIASES: set[str] = {
    "docker.io",
    "index.docker.io",
    _DEFAULT_DOCKER_REGISTRY,
}


@dataclass(frozen=True)
class ParsedImageRef:
    registry: str
    repository: str
    reference: str
    is_digest: bool

    @property
    def full_repository(self) -> str:
        return f"{self.registry}/{self.repository}"


def parse_image_ref(image_ref: str) -> ParsedImageRef:
    name, reference, is_digest = _split_reference(image_ref)
    registry, repository = _split_registry(name)
    registry, repository = _normalize_registry(registry, repository)
    return ParsedImageRef(
        registry=registry,
        repository=repository,
        reference=reference,
        is_digest=is_digest,
    )


def _split_reference(image_ref: str) -> tuple[str, str, bool]:
    if "@" in image_ref:
        name, reference = image_ref.split("@", 1)
        if not reference:
            return image_ref, "", True
        return name, reference, True

    last_colon = image_ref.rfind(":")
    if last_colon > image_ref.rfind("/"):
        name = image_ref[:last_colon]
        reference = image_ref[last_colon + 1 :]
    else:
        name = image_ref
        reference = "latest"
    return name, reference, False


def _split_registry(name: str) -> tuple[str, str]:
    parts = name.split("/", 1)
    if len(parts) == 1:
        return _DEFAULT_DOCKER_REGISTRY, name
    registry, remainder = parts
    if "." in registry or ":" in registry or registry == "localhost":
        return registry, remainder
    return _DEFAULT_DOCKER_REGISTRY, name


def _normalize_registry(registry: str, repository: str) -> tuple[str, str]:
    if registry in _DOCKER_REGISTRY_ALIASES:
        registry = _DEFAULT_DOCKER_REGISTRY
        if "/" not in repository:
            repository = f"library/{repository}"
    return registry, repository



import json
from collections.abc import Callable
from functools import lru_cache
from typing import Any

from aicage.config.errors import ConfigError
from aicage.config.resources import find_packaged_path

_SchemaValidator = Callable[[Any, dict[str, Any], str], None]
_Normalizer = Callable[[dict[str, Any]], dict[str, Any]]


@lru_cache(maxsize=1)
def load_schema(path: str) -> dict[str, Any]:
    payload = find_packaged_path(path).read_text(encoding="utf-8")
    return json.loads(payload)


def validate_schema_mapping(
    mapping: dict[str, Any],
    schema: dict[str, Any],
    context: str,
    *,
    normalizer: _Normalizer | None = None,
    value_validator: _SchemaValidator | None = None,
) -> dict[str, Any]:
    if not isinstance(mapping, dict):
        raise ConfigError(f"{context} must be a mapping.")

    properties = schema.get("properties", {})
    required = set(schema.get("required", []))
    additional = schema.get("additionalProperties", True)

    missing = sorted(required - set(mapping))
    if missing:
        raise ConfigError(f"{context} missing required keys: {', '.join(missing)}.")

    if additional is False:
        unknown = sorted(set(mapping) - set(properties))
        if unknown:
            raise ConfigError(f"{context} contains unsupported keys: {', '.join(unknown)}.")

    normalized = dict(mapping)
    if normalizer is not None:
        normalized = normalizer(normalized)

    if value_validator is not None:
        for key, value in normalized.items():
            schema_entry = properties.get(key)
            if schema_entry is None:
                continue
            value_validator(value, schema_entry, f"{context}.{key}")

    return normalized



from typing import Any


def read_str_list_or_empty(value: Any) -> list[str]:
    if not isinstance(value, list):
        return []
    return [item for item in value if isinstance(item, str) and item]



from ._docker_io import get_docker_io_digest
from ._ghcr import get_ghcr_digest
from ._parser import parse_image_ref


def get_remote_digest(image_ref: str) -> str | None:
    parsed = parse_image_ref(image_ref)
    if parsed.is_digest:
        return parsed.reference
    digest = get_ghcr_digest(parsed)
    if digest:
        return digest
    return get_docker_io_digest(parsed)



from aicage.config.context import ConfigContext
from aicage.config.extensions.loader import ExtensionMetadata
from aicage.config.project_config import AgentConfig
from aicage.registry._errors import RegistryError

from ...runtime.prompts.base import BaseSelectionRequest, prompt_for_base
from ...runtime.prompts.image_choice import ImageChoice, ImageChoiceRequest, prompt_for_image_choice
from ._metadata import available_bases, require_agent_metadata
from .extensions.context import ExtensionSelectionContext
from .extensions.extended_images import (
    apply_extended_selection,
    load_extended_image_options,
    resolve_extended_image,
)
from .extensions.handler import handle_extension_selection
from .models import ImageSelection


def fresh_selection(
    agent: str,
    context: ConfigContext,
    extensions: dict[str, ExtensionMetadata],
) -> ImageSelection:
    agent_metadata = require_agent_metadata(agent, context)
    bases = available_bases(agent, context)
    if not bases:
        raise RegistryError(f"No base images found for agent '{agent}' in config context.")

    extended_images = load_extended_image_options(agent, context)
    request = ImageChoiceRequest(
        agent=agent,
        context=context,
        agent_metadata=agent_metadata,
        extended_options=extended_images,
    )
    choice = (
        prompt_for_image_choice(request)
        if extended_images
        else ImageChoice(kind="base", value=prompt_for_base(_base_request(request)))
    )
    if choice.kind == "extended":
        selected = resolve_extended_image(choice.value, extended_images)
        return apply_extended_selection(
            agent=agent,
            agent_cfg=context.project_cfg.agents.setdefault(agent, AgentConfig()),
            selected=selected,
            agent_metadata=agent_metadata,
            context=context,
        )
    base = choice.value
    agent_cfg = context.project_cfg.agents.setdefault(agent, AgentConfig())
    agent_cfg.base = base
    return handle_extension_selection(
        ExtensionSelectionContext(
            agent=agent,
            base=base,
            agent_cfg=agent_cfg,
            agent_metadata=agent_metadata,
            extensions=extensions,
            context=context,
        )
    )


def _base_request(request: ImageChoiceRequest) -> BaseSelectionRequest:
    return BaseSelectionRequest(
        agent=request.agent,
        context=request.context,
        agent_metadata=request.agent_metadata,
    )



from dataclasses import dataclass
from pathlib import Path

import yaml

from aicage import paths as paths_module
from aicage.registry._sanitize import sanitize

_AGENT_KEY: str = "agent"
_BASE_KEY: str = "base"
_AGENT_VERSION_KEY: str = "agent_version"
_BASE_IMAGE_KEY: str = "base_image"
_IMAGE_REF_KEY: str = "image_ref"
_BUILT_AT_KEY: str = "built_at"


@dataclass(frozen=True)
class BuildRecord:
    agent: str
    base: str
    agent_version: str
    base_image: str
    image_ref: str
    built_at: str


class BuildStore:
    def __init__(self) -> None:
        self._base_dir = paths_module.IMAGE_BUILD_STATE_DIR

    def load(self, agent: str, base: str) -> BuildRecord | None:
        path = self._path(agent, base)
        if not path.is_file():
            return None
        payload = yaml.safe_load(path.read_text(encoding="utf-8")) or {}
        if not isinstance(payload, dict):
            return None
        return BuildRecord(
            agent=str(payload.get(_AGENT_KEY, "")),
            base=str(payload.get(_BASE_KEY, "")),
            agent_version=str(payload.get(_AGENT_VERSION_KEY, "")),
            base_image=str(payload.get(_BASE_IMAGE_KEY, "")),
            image_ref=str(payload.get(_IMAGE_REF_KEY, "")),
            built_at=str(payload.get(_BUILT_AT_KEY, "")),
        )

    def save(self, record: BuildRecord) -> Path:
        self._base_dir.mkdir(parents=True, exist_ok=True)
        path = self._path(record.agent, record.base)
        payload = {
            _AGENT_KEY: record.agent,
            _BASE_KEY: record.base,
            _AGENT_VERSION_KEY: record.agent_version,
            _BASE_IMAGE_KEY: record.base_image,
            _IMAGE_REF_KEY: record.image_ref,
            _BUILT_AT_KEY: record.built_at,
        }
        path.write_text(yaml.safe_dump(payload, sort_keys=True), encoding="utf-8")
        return path

    def _path(self, agent: str, base: str) -> Path:
        filename = f"{sanitize(agent)}-{base}.yaml"
        return self._base_dir / filename



import urllib.error
import urllib.request
from collections.abc import Mapping

from ._timeouts import REGISTRY_REQUEST_TIMEOUT_SECONDS


def head_request(url: str, headers: Mapping[str, str]) -> tuple[int | None, dict[str, str]]:
    request = urllib.request.Request(url, headers=dict(headers), method="HEAD")
    try:
        with urllib.request.urlopen(request, timeout=REGISTRY_REQUEST_TIMEOUT_SECONDS) as response:
            return response.status, dict(response.headers)
    except urllib.error.HTTPError as exc:
        return exc.code, dict(exc.headers)
    except urllib.error.URLError:
        return None, {}


def get_header(headers: Mapping[str, str], key: str) -> str | None:
    for header, value in headers.items():
        if header.lower() == key:
            return value
    return None



from pathlib import Path

from aicage.config.context import ConfigContext
from aicage.config.errors import ConfigError
from aicage.config.yaml_loader import load_yaml
from aicage.registry._errors import RegistryError
from aicage.runtime.prompts.missing_extensions import prompt_for_missing_extensions


def ensure_extensions_exist(
    agent: str,
    context: ConfigContext,
) -> bool:
    agent_cfg = context.project_cfg.agents.get(agent)
    if not agent_cfg:
        return False
    missing = [ext for ext in agent_cfg.extensions if ext not in context.extensions]
    if not missing:
        return False
    other_projects = _find_projects_using_image(context, agent_cfg.image_ref or "")
    choice = prompt_for_missing_extensions(
        agent=agent,
        missing=missing,
        stored_image_ref=agent_cfg.image_ref or "",
        project_config_path=context.store.project_config_path(Path(context.project_cfg.path)),
        other_projects=other_projects,
    )
    if choice == "fresh":
        context.project_cfg.agents.pop(agent, None)
        context.store.save_project(Path(context.project_cfg.path), context.project_cfg)
        return True
    if choice == "exit":
        raise RegistryError("Invalid extension configuration; run aborted.")
    raise RegistryError("Invalid choice; run aborted.")


def _find_projects_using_image(
    context: ConfigContext,
    image_ref: str,
) -> list[tuple[str, Path]]:
    if not image_ref:
        return []
    store = context.store
    matches: list[tuple[str, Path]] = []
    for path in sorted(store.projects_dir.glob("*.yaml")):
        data = _load_yaml(path)
        if not isinstance(data, dict):
            continue
        project_path = str(data.get("path", ""))
        agents = data.get("agents", {}) or {}
        if not isinstance(agents, dict):
            continue
        for cfg in agents.values():
            if isinstance(cfg, dict) and cfg.get("image_ref") == image_ref:
                matches.append((project_path, path))
                break
    return matches


def _load_yaml(path: Path) -> dict[str, object]:
    try:
        return load_yaml(path)
    except ConfigError:
        return {}



from ...constants import DEFAULT_EXTENDED_IMAGE_NAME
from ._tty import ensure_tty_for_prompt


def prompt_for_image_ref(default_ref: str) -> str:
    ensure_tty_for_prompt()
    response = input(f"Enter image name:tag [{default_ref}]: ").strip()
    if not response:
        return default_ref
    if ":" not in response:
        return f"{DEFAULT_EXTENDED_IMAGE_NAME}:{response}"
    return response



from aicage.config.agent.models import AgentMetadata
from aicage.config.context import ConfigContext
from aicage.config.image_refs import local_image_ref
from aicage.constants import IMAGE_REGISTRY, IMAGE_REPOSITORY, LOCAL_IMAGE_REPOSITORY
from aicage.paths import CUSTOM_BASES_DIR


def base_image_ref(
    agent_metadata: AgentMetadata,
    agent: str,
    base: str,
    context: ConfigContext,
) -> str:
    base_metadata = context.bases[base]
    if agent_metadata.build_local or base_metadata.local_definition_dir.is_relative_to(CUSTOM_BASES_DIR):
        return local_image_ref(LOCAL_IMAGE_REPOSITORY, agent, base)
    return local_image_ref(f"{IMAGE_REGISTRY}/{IMAGE_REPOSITORY}", agent, base)



from ._parser import ParsedImageRef
from ._registry import get_manifest_digest

_GHCR_REGISTRY: str = "ghcr.io"


def get_ghcr_digest(parsed: ParsedImageRef) -> str | None:
    if parsed.registry != _GHCR_REGISTRY:
        return None
    return get_manifest_digest(parsed.registry, parsed.repository, parsed.reference)



from pathlib import Path

from aicage.config.project_config import AgentConfig
from aicage.paths import CONTAINER_SSH_DIR, HOST_SSH_DIR
from aicage.runtime.run_args import MountSpec

from ..prompts.confirm import prompt_mount_ssh_keys
from ._signing import is_commit_signing_enabled, resolve_signing_format


def _default_ssh_dir() -> Path:
    return HOST_SSH_DIR


def resolve_ssh_mount(project_path: Path, agent_cfg: AgentConfig) -> list[MountSpec]:
    if not is_commit_signing_enabled(project_path):
        return []
    if resolve_signing_format(project_path) != "ssh":
        return []

    ssh_dir = _default_ssh_dir()
    if not ssh_dir.exists():
        return []

    mounts_cfg = agent_cfg.mounts
    pref = mounts_cfg.ssh
    if pref is None:
        pref = prompt_mount_ssh_keys(ssh_dir)
        mounts_cfg.ssh = pref

    if pref:
        return [
            MountSpec(
                host_path=ssh_dir,
                container_path=CONTAINER_SSH_DIR,
            )
        ]
    return []



from ._parser import ParsedImageRef
from ._registry import get_manifest_digest

_DOCKER_IO_REGISTRY: str = "registry-1.docker.io"


def get_docker_io_digest(parsed: ParsedImageRef) -> str | None:
    if parsed.registry != _DOCKER_IO_REGISTRY:
        return None
    return get_manifest_digest(parsed.registry, parsed.repository, parsed.reference)



from pathlib import Path

from aicage._logging import get_logger
from aicage.config.base.models import BaseMetadata
from aicage.constants import LOCAL_IMAGE_BASE_REPOSITORY
from aicage.docker.build import run_custom_base_build
from aicage.docker.errors import DockerError
from aicage.docker.query import local_image_exists
from aicage.registry._time import now_iso
from aicage.registry.digest.remote_digest import get_remote_digest

from ._custom_base_store import CustomBaseBuildRecord, CustomBaseBuildStore
from ._logs import custom_base_log_path


def custom_base_image_ref(base: str) -> str:
    return f"{LOCAL_IMAGE_BASE_REPOSITORY}:{base}"


def ensure_custom_base_image(base: str, base_metadata: BaseMetadata, base_dir: Path) -> None:
    logger = get_logger()
    image_ref = custom_base_image_ref(base)
    local_exists = local_image_exists(image_ref)
    store = CustomBaseBuildStore()
    record = store.load(base)
    remote_digest = get_remote_digest(base_metadata.from_image)

    if not _should_build(local_exists, record, base_metadata, remote_digest):
        return

    log_path = custom_base_log_path(base)
    try:
        run_custom_base_build(
            dockerfile_path=base_dir / "Dockerfile",
            build_root=base_dir,
            from_image=base_metadata.from_image,
            image_ref=image_ref,
            log_path=log_path,
        )
    except DockerError:
        if local_exists:
            logger.warning(
                "Custom base build failed; using local base image (logs: %s).", log_path
            )
            return
        raise

    store.save(
        CustomBaseBuildRecord(
            base=base,
            from_image=base_metadata.from_image,
            from_image_digest=remote_digest or "",
            image_ref=image_ref,
            built_at=now_iso(),
        )
    )


def _should_build(
    local_exists: bool,
    record: CustomBaseBuildRecord | None,
    base_metadata: BaseMetadata,
    remote_digest: str | None,
) -> bool:
    if not local_exists:
        return True
    if record is None:
        return True
    if record.from_image != base_metadata.from_image:
        return True
    if remote_digest and record.from_image_digest != remote_digest:
        return True
    return False



from pathlib import Path

from .errors import ConfigError


def find_packaged_path(filename: str) -> Path:
    for parent in Path(__file__).resolve().parents:
        candidate = parent / "config" / filename
        if candidate.is_file():
            return candidate
    raise ConfigError(f"Packaged config file '{filename}' was not found.")



from pathlib import Path

from aicage.paths import IMAGE_EXTENDED_BUILD_LOG_DIR
from aicage.registry._sanitize import sanitize
from aicage.registry._time import timestamp


def build_log_path_for_image(image_ref: str) -> Path:
    return IMAGE_EXTENDED_BUILD_LOG_DIR / f"{sanitize(image_ref)}-{timestamp()}.log"



from pathlib import Path
from typing import Any

import yaml

from aicage.config.errors import ConfigError


def load_yaml(path: Path) -> dict[str, Any]:
    try:
        payload = path.read_text(encoding="utf-8")
        data = yaml.safe_load(payload) or {}
    except (OSError, yaml.YAMLError) as exc:
        raise ConfigError(f"Failed to read YAML from {path}: {exc}") from exc
    if not isinstance(data, dict):
        raise ConfigError(f"YAML at {path} must be a mapping.")
    return data



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from aicage._lists import read_str_list_or_empty

_PROJECT_PATH_KEY: str = "path"
_PROJECT_AGENTS_KEY: str = "agents"
_DOCKER_ARGS_KEY: str = "docker_args"

_AGENT_BASE_KEY: str = "base"
_AGENT_ENTRYPOINT_KEY: str = "entrypoint"
_AGENT_MOUNTS_KEY: str = "mounts"
_AGENT_IMAGE_REF_KEY: str = "image_ref"
_AGENT_EXTENSIONS_KEY: str = "extensions"

_MOUNT_GITCONFIG_KEY: str = "gitconfig"
_MOUNT_GNUPG_KEY: str = "gnupg"
_MOUNT_SSH_KEY: str = "ssh"
_MOUNT_DOCKER_KEY: str = "docker"


@dataclass
class _AgentMounts:
    gitconfig: bool | None = None
    gnupg: bool | None = None
    ssh: bool | None = None
    docker: bool | None = None

    @classmethod
    def from_mapping(cls, data: dict[str, Any]) -> "_AgentMounts":
        return cls(
            gitconfig=data.get(_MOUNT_GITCONFIG_KEY),
            gnupg=data.get(_MOUNT_GNUPG_KEY),
            ssh=data.get(_MOUNT_SSH_KEY),
            docker=data.get(_MOUNT_DOCKER_KEY),
        )

    def to_mapping(self) -> dict[str, bool]:
        payload: dict[str, bool] = {}
        if self.gitconfig is not None:
            payload[_MOUNT_GITCONFIG_KEY] = self.gitconfig
        if self.gnupg is not None:
            payload[_MOUNT_GNUPG_KEY] = self.gnupg
        if self.ssh is not None:
            payload[_MOUNT_SSH_KEY] = self.ssh
        if self.docker is not None:
            payload[_MOUNT_DOCKER_KEY] = self.docker
        return payload


@dataclass
class AgentConfig:
    base: str | None = None
    docker_args: str = ""
    entrypoint: str | None = None
    mounts: _AgentMounts = field(default_factory=_AgentMounts)
    image_ref: str | None = None
    extensions: list[str] = field(default_factory=list)

    @classmethod
    def from_mapping(cls, data: dict[str, Any]) -> "AgentConfig":
        mounts = _AgentMounts.from_mapping(data.get(_AGENT_MOUNTS_KEY, {}) or {})
        return cls(
            base=data.get(_AGENT_BASE_KEY),
            docker_args=data.get(_DOCKER_ARGS_KEY, "") or "",
            entrypoint=data.get(_AGENT_ENTRYPOINT_KEY),
            mounts=mounts,
            image_ref=data.get(_AGENT_IMAGE_REF_KEY),
            extensions=read_str_list_or_empty(data.get(_AGENT_EXTENSIONS_KEY)),
        )

    def to_mapping(self) -> dict[str, Any]:
        payload: dict[str, Any] = {}
        if self.base:
            payload[_AGENT_BASE_KEY] = self.base
        if self.docker_args:
            payload[_DOCKER_ARGS_KEY] = self.docker_args
        if self.entrypoint:
            payload[_AGENT_ENTRYPOINT_KEY] = self.entrypoint
        mounts = self.mounts.to_mapping()
        if mounts:
            payload[_AGENT_MOUNTS_KEY] = mounts
        if self.image_ref:
            payload[_AGENT_IMAGE_REF_KEY] = self.image_ref
        if self.extensions:
            payload[_AGENT_EXTENSIONS_KEY] = list(self.extensions)
        return payload


@dataclass
class ProjectConfig:
    path: str
    agents: dict[str, AgentConfig] = field(default_factory=dict)

    @classmethod
    def from_mapping(cls, project_path: Path, data: dict[str, Any]) -> "ProjectConfig":
        raw_agents = data.get(_PROJECT_AGENTS_KEY, {}) or {}
        agents = {name: AgentConfig.from_mapping(cfg) for name, cfg in raw_agents.items()}
        legacy_docker_args = data.get(_DOCKER_ARGS_KEY, "")
        if legacy_docker_args:
            for agent_cfg in agents.values():
                if not agent_cfg.docker_args:
                    agent_cfg.docker_args = legacy_docker_args
        return cls(
            path=data.get(_PROJECT_PATH_KEY, str(project_path)),
            agents=agents,
        )

    def to_mapping(self) -> dict[str, Any]:
        agents_payload = {name: cfg.to_mapping() for name, cfg in self.agents.items()}
        return {_PROJECT_PATH_KEY: self.path, _PROJECT_AGENTS_KEY: agents_payload}



from aicage._logging import get_logger
from aicage.config.runtime_config import RunConfig
from aicage.docker.query import local_image_exists
from aicage.registry._layers import base_layer_missing

from ._extended_store import ExtendedBuildRecord


def should_build_extended(
    run_config: RunConfig,
    record: ExtendedBuildRecord | None,
    base_image_ref: str,
    extension_hash: str,
) -> bool:
    needs_rebuild = (
        not local_image_exists(run_config.selection.image_ref)
        or record is None
        or record.image_ref != run_config.selection.image_ref
        or record.extensions != run_config.selection.extensions
        or record.extension_hash != extension_hash
        or record.base_image != base_image_ref
    )
    if needs_rebuild:
        return True
    is_missing = base_layer_missing(base_image_ref, run_config.selection.image_ref)
    if is_missing is None:
        logger = get_logger()
        logger.warning(
            "Skipping base image layer validation for %s; missing local layer data.",
            run_config.selection.image_ref,
        )
        return False
    return is_missing



IMAGE_REGISTRY: str = "ghcr.io"
_IMAGE_REGISTRY_API_URL: str = "https://ghcr.io/v2"
_IMAGE_REGISTRY_API_TOKEN_URL: str = (
    "https://ghcr.io/token?service=ghcr.io&scope=repository"
)
IMAGE_REPOSITORY: str = "aicage/aicage"
IMAGE_BASE_REPOSITORY: str = "aicage/aicage-image-base"
DEFAULT_IMAGE_BASE: str = "ubuntu"
VERSION_CHECK_IMAGE: str = "ghcr.io/aicage/aicage-image-util:agent-version"

LOCAL_IMAGE_BASE_REPOSITORY: str = "aicage-image-base"
LOCAL_IMAGE_REPOSITORY: str = "aicage"

DEFAULT_EXTENDED_IMAGE_NAME: str = "aicage-extended"



from aicage._logging import get_logger
from aicage.docker.pull import run_pull
from aicage.docker.query import get_local_repo_digest_for_repo
from aicage.registry._errors import RegistryError
from aicage.registry._logs import pull_log_path
from aicage.registry.digest.remote_digest import get_remote_digest


def refresh_base_digest(
    base_image_ref: str,
    base_repository: str,
) -> str | None:
    logger = get_logger()
    local_digest = get_local_repo_digest_for_repo(base_image_ref, base_repository)
    remote_digest = get_remote_digest(base_image_ref)
    if remote_digest is None or remote_digest == local_digest:
        return local_digest

    log_path = pull_log_path(base_image_ref)
    try:
        run_pull(base_image_ref, log_path)
    except RegistryError:
        if local_digest:
            logger.warning(
                "Base image pull failed; using local base image (logs: %s).", log_path
            )
            return local_digest
        raise

    return get_local_repo_digest_for_repo(base_image_ref, base_repository)



from aicage.constants import IMAGE_REGISTRY, IMAGE_REPOSITORY
from aicage.docker.query import get_local_repo_digest
from aicage.docker.types import ImageRefRepository
from aicage.registry.digest.remote_digest import get_remote_digest


def decide_pull(image_ref: str) -> bool:
    # Local digests include registry prefix; registry API uses repository only.
    local_repository = f"{IMAGE_REGISTRY}/{IMAGE_REPOSITORY}"
    local_digest = get_local_repo_digest(
        ImageRefRepository(
            image_ref=image_ref,
            repository=local_repository,
        )
    )
    if local_digest is None:
        return True

    remote_digest = get_remote_digest(image_ref)
    if remote_digest is None:
        return False

    return local_digest != remote_digest



from pathlib import Path

from ..paths import IMAGE_PULL_LOG_DIR
from ._sanitize import sanitize
from ._time import timestamp


def pull_log_path(image_ref: str) -> Path:
    return IMAGE_PULL_LOG_DIR / f"{sanitize(image_ref)}-{timestamp()}.log"



import os
import subprocess
from dataclasses import dataclass
from pathlib import Path

from aicage._logging import get_logger
from aicage.docker.run import run_builder_version_check


@dataclass(frozen=True)
class _CommandResult:
    success: bool
    output: str
    error: str


def run_version_check_image(image_ref: str, definition_dir: Path) -> _CommandResult:
    process = run_builder_version_check(image_ref, definition_dir)
    return _from_process(process, "version check image")


def run_host(script_path: Path) -> _CommandResult:
    if not os.access(script_path, os.X_OK):
        get_logger().warning(
            "version.sh at %s is not executable; running with /bin/bash.",
            script_path,
        )
    return _run_command(["bash", str(script_path)], "host")


def _run_command(command: list[str], context: str) -> _CommandResult:
    try:
        process = subprocess.run(command, check=False, capture_output=True, text=True)
    except Exception as exc:
        get_logger().warning("Version check failed in %s: %s", context, exc)
        return _CommandResult(success=False, output="", error=str(exc))
    return _from_process(process, context)


def _from_process(process: subprocess.CompletedProcess[str], context: str) -> _CommandResult:
    output = process.stdout.strip() if process.stdout else ""
    if process.returncode == 0 and output:
        return _CommandResult(success=True, output=output, error="")

    stderr = process.stderr.strip() if process.stderr else ""
    error = stderr or output or f"Version check failed in {context}."
    return _CommandResult(success=False, output=output, error=error)



from dataclasses import dataclass

from aicage.config.agent.models import AgentMetadata
from aicage.config.base.models import BaseMetadata
from aicage.config.extensions.loader import ExtensionMetadata
from aicage.constants import IMAGE_REGISTRY, IMAGE_REPOSITORY

from .config_store import SettingsStore
from .project_config import ProjectConfig


@dataclass
class ConfigContext:
    store: SettingsStore
    project_cfg: ProjectConfig
    agents: dict[str, AgentMetadata]
    bases: dict[str, BaseMetadata]
    extensions: dict[str, ExtensionMetadata]

    @staticmethod
    def image_repository_ref() -> str:
        return f"{IMAGE_REGISTRY}/{IMAGE_REPOSITORY}"



from dataclasses import dataclass, field
from pathlib import Path

AGENT_PATH_KEY: str = "agent_path"
AGENT_FULL_NAME_KEY: str = "agent_full_name"
AGENT_HOMEPAGE_KEY: str = "agent_homepage"
BUILD_LOCAL_KEY: str = "build_local"
BASE_EXCLUDE_KEY: str = "base_exclude"
BASE_DISTRO_EXCLUDE_KEY: str = "base_distro_exclude"


@dataclass(frozen=True)
class AgentMetadata:
    agent_path: str
    agent_full_name: str
    agent_homepage: str
    build_local: bool
    valid_bases: dict[str, str]
    local_definition_dir: Path
    base_exclude: list[str] = field(default_factory=list)
    base_distro_exclude: list[str] = field(default_factory=list)



from typing import Any

from aicage.config.errors import ConfigError


def expect_string(value: Any, context: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise ConfigError(f"{context} must be a non-empty string.")
    return value


def expect_bool(value: Any, context: str) -> bool:
    if not isinstance(value, bool):
        raise ConfigError(f"{context} must be a boolean.")
    return value


def read_str_list(value: Any, context: str) -> list[str]:
    if not isinstance(value, list):
        raise ConfigError(f"{context} must be a list.")
    items: list[str] = []
    for item in value:
        if not isinstance(item, str) or not item.strip():
            raise ConfigError(f"{context} must contain non-empty strings.")
        items.append(item)
    return items


def maybe_str_list(value: Any, context: str) -> list[str] | None:
    if value is None:
        return None
    return read_str_list(value, context)


def expect_keys(
    mapping: dict[str, Any],
    required: set[str],
    optional: set[str],
    context: str,
) -> None:
    missing = sorted(required - set(mapping))
    if missing:
        raise ConfigError(f"{context} missing required keys: {', '.join(missing)}.")
    unknown = sorted(set(mapping) - required - optional)
    if unknown:
        raise ConfigError(f"{context} contains unsupported keys: {', '.join(unknown)}.")



from pathlib import Path

from aicage.config._yaml import expect_bool, expect_string
from aicage.config.base._custom_loader import load_custom_bases
from aicage.config.base._validation import validate_base_mapping
from aicage.config.base.models import BUILD_LOCAL_KEY, BaseMetadata
from aicage.config.errors import ConfigError
from aicage.config.resources import find_packaged_path
from aicage.config.yaml_loader import load_yaml

_BASES_DIR_NAME: str = "base-build/bases"
_BASE_DEFINITION_FILES: tuple[str, str] = ("base.yaml", "base.yml")
_FROM_IMAGE_KEY: str = "from_image"
_BASE_IMAGE_DISTRO_KEY: str = "base_image_distro"
_BASE_IMAGE_DESCRIPTION_KEY: str = "base_image_description"


def load_bases() -> dict[str, BaseMetadata]:
    builtin_bases = _load_builtin_bases()
    custom_bases = load_custom_bases()
    merged_bases = dict(builtin_bases)
    merged_bases.update(custom_bases)
    return merged_bases


def _load_builtin_bases() -> dict[str, BaseMetadata]:
    bases_dir = _builtin_bases_dir()
    if not bases_dir.is_dir():
        raise ConfigError(f"Built-in base directory '{bases_dir}' is missing.")

    bases: dict[str, BaseMetadata] = {}
    for entry in sorted(bases_dir.iterdir()):
        if not entry.is_dir():
            continue
        base_name = entry.name
        definition_path = _find_base_definition(entry)
        mapping = validate_base_mapping(load_yaml(definition_path))
        bases[base_name] = BaseMetadata(
            from_image=expect_string(mapping.get(_FROM_IMAGE_KEY), _FROM_IMAGE_KEY),
            base_image_distro=expect_string(mapping.get(_BASE_IMAGE_DISTRO_KEY), _BASE_IMAGE_DISTRO_KEY),
            base_image_description=expect_string(
                mapping.get(_BASE_IMAGE_DESCRIPTION_KEY),
                _BASE_IMAGE_DESCRIPTION_KEY,
            ),
            build_local=expect_bool(mapping.get(BUILD_LOCAL_KEY), BUILD_LOCAL_KEY),
            local_definition_dir=entry,
        )
    return bases


def _builtin_bases_dir() -> Path:
    dockerfile = find_packaged_path("agent-build/Dockerfile")
    return dockerfile.parent.parent / _BASES_DIR_NAME


def _find_base_definition(base_dir: Path) -> Path:
    for filename in _BASE_DEFINITION_FILES:
        candidate = base_dir / filename
        if candidate.is_file():
            return candidate
    expected = ", ".join(_BASE_DEFINITION_FILES)
    raise ConfigError(f"Base '{base_dir.name}' is missing {expected}.")



"""Shared environment variable names used for container setup."""

AICAGE_UID = "AICAGE_UID"
AICAGE_GID = "AICAGE_GID"
AICAGE_USER = "AICAGE_USER"
AICAGE_WORKSPACE = "AICAGE_WORKSPACE"

AICAGE_AGENT_CONFIG_PATH = "AICAGE_AGENT_CONFIG_PATH"




from pathlib import Path

from aicage._logging import get_logger
from aicage.config.agent.models import AgentMetadata
from aicage.constants import VERSION_CHECK_IMAGE
from aicage.registry._errors import RegistryError

from ._command import run_host, run_version_check_image
from ._images import ensure_version_check_image
from ._store import VersionCheckStore


class AgentVersionChecker:
    def __init__(self, store: VersionCheckStore | None = None) -> None:
        self._store = store or VersionCheckStore()

    def get_version(
        self,
        agent_name: str,
        _agent_metadata: AgentMetadata,
        definition_dir: Path,
    ) -> str:
        logger = get_logger()
        script_path = definition_dir / "version.sh"
        if not script_path.is_file():
            raise RegistryError(f"Agent '{agent_name}' is missing version.sh at {script_path}.")

        errors: list[str] = []
        host_result = run_host(script_path)
        if host_result.success:
            logger.info("Version check succeeded on host for %s", agent_name)
            self._store.save(agent_name, host_result.output)
            return host_result.output

        logger.warning(
            "Version check failed on host for %s: %s",
            agent_name,
            host_result.error,
        )
        errors.append(host_result.error)

        ensure_version_check_image(VERSION_CHECK_IMAGE, logger)
        image_result = run_version_check_image(VERSION_CHECK_IMAGE, definition_dir)
        if image_result.success:
            logger.info("Version check succeeded in version check image for %s", agent_name)
            self._store.save(agent_name, image_result.output)
            return image_result.output

        logger.warning(
            "Version check failed in version check image for %s: %s",
            agent_name,
            image_result.error,
        )
        errors.append(image_result.error)
        logger.error("Version check failed for %s: %s", agent_name, "; ".join(errors))
        raise RegistryError("; ".join(errors))



import hashlib

from aicage.config.extensions.loader import ExtensionMetadata, extension_hash
from aicage.config.runtime_config import RunConfig
from aicage.docker.build import run_extended_build
from aicage.registry._errors import RegistryError
from aicage.registry._time import now_iso

from ._extended_plan import should_build_extended
from ._extended_store import ExtendedBuildRecord, ExtendedBuildStore
from ._logs import build_log_path_for_image


def ensure_extended_image(run_config: RunConfig) -> None:
    if not run_config.selection.extensions:
        raise RegistryError("No extensions selected for extended image build.")

    resolved = _resolve_extensions(run_config.selection.extensions, run_config.context.extensions)
    combined_hash = _combined_extension_hash(resolved)
    store = ExtendedBuildStore()
    record = store.load(run_config.selection.image_ref)
    needs_build = should_build_extended(
        run_config=run_config,
        record=record,
        base_image_ref=run_config.selection.base_image_ref,
        extension_hash=combined_hash,
    )
    if not needs_build:
        return

    log_path = build_log_path_for_image(run_config.selection.image_ref)
    run_extended_build(
        run_config=run_config,
        base_image_ref=run_config.selection.base_image_ref,
        extensions=resolved,
        log_path=log_path,
    )
    store.save(
        ExtendedBuildRecord(
            agent=run_config.agent,
            base=run_config.selection.base,
            image_ref=run_config.selection.image_ref,
            extensions=list(run_config.selection.extensions),
            extension_hash=combined_hash,
            base_image=run_config.selection.base_image_ref,
            built_at=now_iso(),
        )
    )


def _resolve_extensions(
    extension_ids: list[str],
    extensions: dict[str, ExtensionMetadata],
) -> list[ExtensionMetadata]:
    missing = [ext for ext in extension_ids if ext not in extensions]
    if missing:
        raise RegistryError(f"Missing extensions: {', '.join(sorted(missing))}.")
    return [extensions[ext] for ext in extension_ids]


def _combined_extension_hash(extensions: list[ExtensionMetadata]) -> str:
    digest = hashlib.sha256()
    for extension in extensions:
        digest.update(extension.extension_id.encode("utf-8"))
        digest.update(extension_hash(extension).encode("utf-8"))
    return digest.hexdigest()



import os
from os.path import expanduser
from pathlib import Path, PurePosixPath, PureWindowsPath

# AICAGE PATHS (for Python program on host)

_AGENT_DEFINITION_FILENAME: str = "agent.yaml"
EXTENDED_IMAGE_DEFINITION_FILENAME: str = "image-extended.yaml"

_CONFIG_BASE_DIR: Path = Path(expanduser("~/.aicage"))
PROJECTS_DIR: Path = _CONFIG_BASE_DIR / "projects"

BASE_IMAGE_BUILD_STATE_DIR: Path = _CONFIG_BASE_DIR / "state/base-image/build"
IMAGE_BUILD_STATE_DIR: Path = _CONFIG_BASE_DIR / "state/image/build"
AGENT_VERSION_CHECK_STATE_DIR: Path = _CONFIG_BASE_DIR / "state/agent/version-check/state"
IMAGE_EXTENDED_STATE_DIR: Path = _CONFIG_BASE_DIR / "state/image-extended/state"
IMAGE_EXTENDED_BUILD_STATE_DIR: Path =  _CONFIG_BASE_DIR / "state/image-extended/build"

_LOG_DIR: Path = _CONFIG_BASE_DIR / "logs"
GLOBAL_LOG_PATH: Path = _LOG_DIR / "aicage.log"
IMAGE_PULL_LOG_DIR: Path = _LOG_DIR / "image/pull"
BASE_IMAGE_BUILD_LOG_DIR: Path = _LOG_DIR / "base-image/build"
IMAGE_BUILD_LOG_DIR: Path = _LOG_DIR / "image/build"
IMAGE_EXTENDED_BUILD_LOG_DIR: Path = _LOG_DIR / "image-extended/build"

# Only user-generated custom files outside ~/.aicage.
_CUSTOM_ROOT_DIR: Path = Path(expanduser("~/.aicage-custom"))

CUSTOM_BASES_DIR: Path = _CUSTOM_ROOT_DIR / "base-images"
CUSTOM_BASE_DEFINITION_FILES: tuple[str, str] = (
    "base.yaml",
    "base.yml",
)

CUSTOM_AGENTS_DIR: Path = _CUSTOM_ROOT_DIR / "agents"
CUSTOM_AGENT_DEFINITION_FILES: tuple[str, str] = (
    _AGENT_DEFINITION_FILENAME,
    "agent.yml",
)

CUSTOM_EXTENSIONS_DIR: Path = _CUSTOM_ROOT_DIR / "extensions"
CUSTOM_EXTENSION_DEFINITION_FILES: tuple[str, str] = (
    "extension.yaml",
    "extension.yml",
)

# OTHER HOST PATHS (for mounts to container)

HOST_SSH_DIR: Path = Path.home() / ".ssh"
HOST_DOCKER_SOCKET_PATH: Path = Path("/run/docker.sock")

# CONTAINER PATHS (for mounts to container)

CONTAINER_WORKSPACE_DIR: PurePosixPath = PurePosixPath("/workspace")
_WSL_MOUNT_ROOT: PurePosixPath = PurePosixPath("/mnt")
CONTAINER_AGENT_CONFIG_DIR: PurePosixPath = PurePosixPath("/aicage/agent-config")
CONTAINER_GITCONFIG_PATH: PurePosixPath = PurePosixPath("/aicage/host/gitconfig")
CONTAINER_SSH_DIR: PurePosixPath = PurePosixPath("/aicage/host/ssh")
CONTAINER_GNUPG_DIR: PurePosixPath = PurePosixPath("/aicage/host/gnupg")
CONTAINER_ENTRYPOINT_PATH: PurePosixPath = PurePosixPath("/usr/local/bin/entrypoint.sh")
CONTAINER_DOCKER_SOCKET_PATH: PurePosixPath = PurePosixPath("/run/docker.sock")


def container_project_path(host_path: Path) -> PurePosixPath:
    if os.name != "nt":
        return PurePosixPath(host_path.as_posix())
    win_path = PureWindowsPath(host_path)
    drive = win_path.drive
    if drive.startswith("\\\\?\\"):
        drive = drive[4:]
    drive_letter = drive.rstrip(":").lower()
    if not drive_letter:
        return PurePosixPath(host_path.as_posix())
    parts = [part for part in win_path.parts if part != win_path.anchor]
    return _WSL_MOUNT_ROOT / drive_letter / PurePosixPath(*parts)



from aicage.config.agent.models import AgentMetadata
from aicage.config.context import ConfigContext


def filter_bases(context: ConfigContext, agent_metadata: AgentMetadata) -> set[str]:
    base_exclude = _normalize_exclude(agent_metadata.base_exclude)
    base_distro_exclude = _normalize_exclude(agent_metadata.base_distro_exclude)
    filtered: set[str] = set()
    for base_name, base_metadata in context.bases.items():
        if _is_base_excluded(
            base_name,
            base_metadata.base_image_distro,
            base_exclude,
            base_distro_exclude,
        ):
            continue
        filtered.add(base_name)
    return filtered


def _is_base_excluded(
    base_name: str,
    base_distro: str,
    base_exclude: set[str],
    base_distro_exclude: set[str],
) -> bool:
    base_name_lc = base_name.lower()
    if base_name_lc in base_exclude:
        return True
    if base_distro.lower() in base_distro_exclude:
        return True
    return False


def _normalize_exclude(values: list[str]) -> set[str]:
    if not values:
        return set()
    return {value.lower() for value in values}



from pathlib import Path

from aicage.config.agent._metadata import build_agent_metadata
from aicage.config.agent._validation import ensure_required_files
from aicage.config.agent.models import AgentMetadata
from aicage.config.base.models import BaseMetadata
from aicage.config.errors import ConfigError
from aicage.config.yaml_loader import load_yaml
from aicage.paths import CUSTOM_AGENT_DEFINITION_FILES, CUSTOM_AGENTS_DIR


def load_custom_agents(
    bases: dict[str, BaseMetadata],
) -> dict[str, AgentMetadata]:
    agents_dir = CUSTOM_AGENTS_DIR
    if not agents_dir.is_dir():
        return {}

    custom_agents: dict[str, AgentMetadata] = {}
    for entry in sorted(agents_dir.iterdir()):
        if not entry.is_dir():
            continue
        agent_name = entry.name
        agent_path = _find_agent_definition(entry)
        agent_mapping = load_yaml(agent_path)
        ensure_required_files(agent_name, entry)
        custom_agents[agent_name] = build_agent_metadata(
            agent_name=agent_name,
            agent_mapping=agent_mapping,
            bases=bases,
            definition_dir=entry,
        )
    return custom_agents


def _find_agent_definition(agent_dir: Path) -> Path:
    for filename in CUSTOM_AGENT_DEFINITION_FILES:
        candidate = agent_dir / filename
        if candidate.is_file():
            return candidate
    expected = ", ".join(CUSTOM_AGENT_DEFINITION_FILES)
    raise ConfigError(f"Custom agent '{agent_dir.name}' is missing {expected}.")



from aicage._logging import get_logger
from aicage.config.runtime_config import RunConfig
from aicage.docker.query import local_image_exists
from aicage.registry._layers import base_layer_missing

from ._refs import get_base_image_ref
from ._store import BuildRecord


def should_build(
    run_config: RunConfig,
    record: BuildRecord | None,
    agent_version: str,
) -> bool:
    base_image_ref = get_base_image_ref(run_config)
    image_ref = run_config.selection.base_image_ref
    if not local_image_exists(image_ref):
        return True
    if record is None:
        return True
    if record.agent_version != agent_version:
        return True
    is_missing = base_layer_missing(base_image_ref, image_ref)
    if is_missing is None:
        logger = get_logger()
        logger.warning(
            "Skipping base image layer validation for %s; missing local layer data.",
            image_ref,
        )
        return False
    if is_missing:
        return True
    return False



from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig

from ._fresh_selection import fresh_selection
from ._metadata import require_agent_metadata, validate_base
from .extensions.context import ExtensionSelectionContext
from .extensions.handler import handle_extension_selection
from .extensions.missing_extensions import ensure_extensions_exist
from .extensions.refs import base_image_ref
from .models import ImageSelection


def select_agent_image(agent: str, context: ConfigContext) -> ImageSelection:
    extensions = context.extensions
    agent_cfg = context.project_cfg.agents.setdefault(agent, AgentConfig())
    agent_metadata = require_agent_metadata(agent, context)
    base = agent_cfg.base

    if agent_cfg.image_ref:
        if base is None:
            return fresh_selection(agent, context, extensions)
        validate_base(agent, base, context)
        if agent_cfg.extensions:
            reset = ensure_extensions_exist(
                context=context,
                agent=agent,
            )
            if reset:
                return fresh_selection(agent, context, extensions)
        return ImageSelection(
            image_ref=agent_cfg.image_ref,
            base=base,
            extensions=list(agent_cfg.extensions),
            base_image_ref=base_image_ref(agent_metadata, agent, base, context),
        )

    if not base:
        return fresh_selection(agent, context, extensions)

    validate_base(agent, base, context)
    return handle_extension_selection(
        ExtensionSelectionContext(
            agent=agent,
            base=base,
            agent_cfg=agent_cfg,
            agent_metadata=agent_metadata,
            extensions=extensions,
            context=context,
        )
    )



import os
from pathlib import Path

from aicage.config.project_config import AgentConfig
from aicage.paths import CONTAINER_ENTRYPOINT_PATH
from aicage.runtime._errors import RuntimeExecutionError
from aicage.runtime.prompts.confirm import prompt_persist_entrypoint
from aicage.runtime.run_args import MountSpec


def resolve_entrypoint_mount(
    agent_cfg: AgentConfig,
    cli_entrypoint: str | None,
) -> list[MountSpec]:
    entrypoint_value = cli_entrypoint or agent_cfg.entrypoint
    if not entrypoint_value:
        return []

    entrypoint_path = _resolve_entrypoint_path(entrypoint_value)
    _validate_entrypoint_path(entrypoint_path)
    mounts = [
        MountSpec(
            host_path=entrypoint_path,
            container_path=CONTAINER_ENTRYPOINT_PATH,
            read_only=True,
        )
    ]

    if cli_entrypoint and agent_cfg.entrypoint is None:
        if prompt_persist_entrypoint(entrypoint_path):
            agent_cfg.entrypoint = str(entrypoint_path)

    return mounts


def _resolve_entrypoint_path(entrypoint: str) -> Path:
    return Path(entrypoint).expanduser().resolve()


def _validate_entrypoint_path(path: Path) -> None:
    if not path.exists() or not path.is_file():
        raise RuntimeExecutionError(f"Entrypoint '{path}' does not exist or is not a file.")
    if os.name != "nt" and not os.access(path, os.X_OK):
        raise RuntimeExecutionError(f"Entrypoint '{path}' is not executable.")



from pathlib import Path

from aicage.paths import BASE_IMAGE_BUILD_LOG_DIR, IMAGE_BUILD_LOG_DIR
from aicage.registry._sanitize import sanitize
from aicage.registry._time import timestamp


def build_log_path(agent: str, base: str) -> Path:
    return IMAGE_BUILD_LOG_DIR / f"{sanitize(agent)}-{base}-{timestamp()}.log"


def custom_base_log_path(base: str) -> Path:
    return BASE_IMAGE_BUILD_LOG_DIR / f"{sanitize(base)}-{timestamp()}.log"



from pathlib import Path

from aicage.config._yaml import expect_string
from aicage.config.base._validation import validate_base_mapping
from aicage.config.base.models import BaseMetadata
from aicage.config.errors import ConfigError
from aicage.config.yaml_loader import load_yaml
from aicage.paths import CUSTOM_BASE_DEFINITION_FILES, CUSTOM_BASES_DIR

_FROM_IMAGE_KEY: str = "from_image"
_BASE_IMAGE_DISTRO_KEY: str = "base_image_distro"
_BASE_IMAGE_DESCRIPTION_KEY: str = "base_image_description"

_DOCKERFILE_NAME: str = "Dockerfile"


def load_custom_bases() -> dict[str, BaseMetadata]:
    bases_dir = CUSTOM_BASES_DIR
    if not bases_dir.is_dir():
        return {}

    custom_bases: dict[str, BaseMetadata] = {}
    for entry in sorted(bases_dir.iterdir()):
        if not entry.is_dir():
            continue
        base_name = entry.name
        custom_bases[base_name] = _load_custom_base(base_name)
    return custom_bases


def _load_custom_base(base_name: str) -> BaseMetadata:
    base_dir = CUSTOM_BASES_DIR / base_name
    definition_path = _find_base_definition(base_dir)
    _ensure_required_files(base_name, base_dir)
    mapping = validate_base_mapping(load_yaml(definition_path))
    return BaseMetadata(
        from_image=expect_string(mapping.get(_FROM_IMAGE_KEY), _FROM_IMAGE_KEY),
        base_image_distro=expect_string(mapping.get(_BASE_IMAGE_DISTRO_KEY), _BASE_IMAGE_DISTRO_KEY),
        base_image_description=expect_string(
            mapping.get(_BASE_IMAGE_DESCRIPTION_KEY),
            _BASE_IMAGE_DESCRIPTION_KEY,
        ),
        build_local=True,
        local_definition_dir=base_dir,
    )


def _find_base_definition(base_dir: Path) -> Path:
    for filename in CUSTOM_BASE_DEFINITION_FILES:
        candidate = base_dir / filename
        if candidate.is_file():
            return candidate
    expected = ", ".join(CUSTOM_BASE_DEFINITION_FILES)
    raise ConfigError(f"Custom base '{base_dir.name}' is missing {expected}.")


def _ensure_required_files(base_name: str, base_dir: Path) -> None:
    dockerfile = base_dir / _DOCKERFILE_NAME
    if not dockerfile.is_file():
        raise ConfigError(f"Custom base '{base_name}' is missing {_DOCKERFILE_NAME}.")



from dataclasses import dataclass


@dataclass(frozen=True)
class ImageRefRepository:
    image_ref: str
    repository: str


@dataclass(frozen=True)
class RegistryApiConfig:
    registry_api_url: str
    registry_api_token_url: str




from pathlib import Path

from aicage.config.project_config import AgentConfig
from aicage.paths import CONTAINER_GNUPG_DIR
from aicage.runtime.run_args import MountSpec

from ..prompts.confirm import prompt_mount_gpg_keys
from ._exec import capture_stdout
from ._signing import is_commit_signing_enabled, resolve_signing_format


def _resolve_gpg_home() -> Path | None:
    stdout = capture_stdout(["gpgconf", "--list-dirs", "homedir"])
    if stdout:
        path = stdout.strip()
        if path:
            gpg_home = Path(path).expanduser()
            if gpg_home.exists():
                return gpg_home
    fallback = Path("~/.gnupg").expanduser()
    return fallback if fallback.exists() else None


def resolve_gpg_mount(project_path: Path, agent_cfg: AgentConfig) -> list[MountSpec]:
    if not is_commit_signing_enabled(project_path):
        return []
    if resolve_signing_format(project_path) == "ssh":
        return []

    gpg_home = _resolve_gpg_home()
    if not gpg_home or not gpg_home.exists():
        return []

    mounts_cfg = agent_cfg.mounts
    pref = mounts_cfg.gnupg
    if pref is None:
        pref = prompt_mount_gpg_keys(gpg_home)
        mounts_cfg.gnupg = pref

    if pref:
        return [
            MountSpec(
                host_path=gpg_home,
                container_path=CONTAINER_GNUPG_DIR,
            )
        ]
    return []



from aicage.config.agent.models import AgentMetadata
from aicage.config.base.filter import filter_bases
from aicage.config.context import ConfigContext
from aicage.registry._errors import RegistryError


def require_agent_metadata(agent: str, context: ConfigContext) -> AgentMetadata:
    agent_metadata = context.agents.get(agent)
    if not agent_metadata:
        raise RegistryError(f"Agent '{agent}' is missing from config context.")
    return agent_metadata


def available_bases(
    agent: str,
    context: ConfigContext,
) -> list[str]:
    agent_metadata = require_agent_metadata(agent, context)
    filtered = filter_bases(context, agent_metadata)
    if not filtered:
        raise RegistryError(f"Agent '{agent}' does not define any valid bases.")
    return sorted(filtered)


def validate_base(
    agent: str,
    base: str,
    context: ConfigContext,
) -> None:
    agent_metadata = require_agent_metadata(agent, context)
    if base not in filter_bases(context, agent_metadata):
        raise RegistryError(f"Base '{base}' is not valid for agent '{agent}'.")



from aicage.config.runtime_config import RunConfig
from aicage.paths import CUSTOM_BASES_DIR
from aicage.registry._image_pull import pull_image
from aicage.registry.extension_build.ensure_extended_image import ensure_extended_image
from aicage.registry.local_build.ensure_local_image import ensure_local_image


def ensure_image(run_config: RunConfig) -> None:
    agent_metadata = run_config.context.agents[run_config.agent]
    base_metadata = run_config.context.bases[run_config.selection.base]
    custom_base = base_metadata.local_definition_dir.is_relative_to(CUSTOM_BASES_DIR)
    if not agent_metadata.build_local and not custom_base:
        pull_image(run_config.selection.base_image_ref)
    else:
        ensure_local_image(run_config)
    if run_config.selection.extensions:
        ensure_extended_image(run_config)



from dataclasses import dataclass

from aicage._logging import get_logger
from aicage.config.agent.models import AgentMetadata
from aicage.config.context import ConfigContext
from aicage.constants import DEFAULT_IMAGE_BASE
from aicage.runtime._errors import RuntimeExecutionError

from ._tty import ensure_tty_for_prompt
from .base import BaseOption, available_bases, base_options


@dataclass(frozen=True)
class ExtendedImageOption:
    name: str
    base: str
    description: str
    extensions: list[str]
    image_ref: str


@dataclass(frozen=True)
class ImageChoiceRequest:
    agent: str
    context: ConfigContext
    agent_metadata: AgentMetadata
    extended_options: list[ExtendedImageOption]


@dataclass(frozen=True)
class ImageChoice:
    kind: str
    value: str


def prompt_for_image_choice(request: ImageChoiceRequest) -> ImageChoice:
    ensure_tty_for_prompt()
    logger = get_logger()
    bases = base_options(request.context, request.agent_metadata)
    extended = request.extended_options
    options = _build_image_options(bases, extended)
    prompt = _render_image_prompt(request, options)
    response = input(prompt).strip()
    choice = _parse_image_choice_response(response, bases, extended, options)
    logger.info("Selected %s '%s' for agent '%s'", choice.kind, choice.value, request.agent)
    return choice


def _build_image_options(
    bases: list[BaseOption],
    extended: list[ExtendedImageOption],
) -> list[tuple[str, ImageChoice]]:
    options: list[tuple[str, ImageChoice]] = []
    for option in bases:
        label = f"{option.base}: {option.description}"
        options.append((label, ImageChoice(kind="base", value=option.base)))
    for option in extended:
        extension_list = ", ".join(option.extensions)
        label = (
            f"{option.name}: {option.description} (base {option.base}, extensions: {extension_list})"
        )
        options.append((label, ImageChoice(kind="extended", value=option.name)))
    return options


def _render_image_prompt(
    request: ImageChoiceRequest,
    options: list[tuple[str, ImageChoice]],
) -> str:
    title = f"Select image for '{request.agent}' (runtime to use inside the container):"
    if not options:
        return f"{title} [{DEFAULT_IMAGE_BASE}]: "
    print(title)
    for idx, (label, choice) in enumerate(options, start=1):
        suffix = ""
        if choice.kind == "base" and choice.value == DEFAULT_IMAGE_BASE:
            suffix = " (default)"
        print(f"  {idx}) {label}{suffix}")
    return f"Enter number or name [{DEFAULT_IMAGE_BASE}]: "


def _parse_image_choice_response(
    response: str,
    bases: list[BaseOption],
    extended: list[ExtendedImageOption],
    options: list[tuple[str, ImageChoice]],
) -> ImageChoice:
    if not response:
        return ImageChoice(kind="base", value=DEFAULT_IMAGE_BASE)
    if response.isdigit() and options:
        idx = int(response)
        if idx < 1 or idx > len(options):
            raise RuntimeExecutionError(
                f"Invalid choice '{response}'. Pick a number between 1 and {len(options)}."
            )
        return options[idx - 1][1]
    base_match = _match_base_choice(response, bases)
    if base_match is not None:
        return ImageChoice(kind="base", value=base_match)
    extended_match = _match_extended_choice(response, extended)
    if extended_match is None:
        valid = ", ".join(available_bases(bases) + [option.name for option in extended])
        raise RuntimeExecutionError(f"Invalid choice '{response}'. Valid options: {valid}")
    return ImageChoice(kind="extended", value=extended_match)


def _match_base_choice(response: str, options: list[BaseOption]) -> str | None:
    if response in available_bases(options):
        return response
    return None


def _match_extended_choice(response: str, options: list[ExtendedImageOption]) -> str | None:
    for option in options:
        if option.name == response:
            return option.name
    return None



from logging import Logger

from aicage.constants import IMAGE_REGISTRY
from aicage.docker.pull import run_pull
from aicage.docker.query import get_local_repo_digest_for_repo
from aicage.registry._errors import RegistryError
from aicage.registry._logs import pull_log_path
from aicage.registry.digest.remote_digest import get_remote_digest


def ensure_version_check_image(image_ref: str, logger: Logger) -> None:
    local_digest = get_local_repo_digest_for_repo(
        image_ref,
        _local_repository(image_ref, IMAGE_REGISTRY),
    )
    if local_digest is None:
        _pull_version_check_image(image_ref, logger)
        return

    remote_digest = get_remote_digest(image_ref)
    if remote_digest is None or remote_digest == local_digest:
        return

    _pull_version_check_image(image_ref, logger)


def _pull_version_check_image(image_ref: str, logger: Logger) -> None:
    log_path = pull_log_path(image_ref)
    try:
        run_pull(image_ref, log_path)
    except RegistryError:
        logger.warning("Version check image pull failed; using local image (logs: %s).", log_path)


def _local_repository(image_ref: str, default_registry: str) -> str:
    name = _strip_reference(image_ref)
    parts = name.split("/", 1)
    if len(parts) == 1:
        return f"{default_registry}/{name}"
    registry, remainder = parts
    if "." in registry or ":" in registry or registry == "localhost":
        return f"{registry}/{remainder}"
    return f"{default_registry}/{name}"


def _strip_reference(image_ref: str) -> str:
    if "@" in image_ref:
        return image_ref.split("@", 1)[0]
    last_colon = image_ref.rfind(":")
    if last_colon > image_ref.rfind("/"):
        return image_ref[:last_colon]
    return image_ref



from collections.abc import Mapping

from ._auth import fetch_bearer_token, parse_auth_header
from ._http import get_header, head_request

_ACCEPT_HEADERS = ",".join(
    [
        "application/vnd.oci.image.index.v1+json",
        "application/vnd.docker.distribution.manifest.list.v2+json",
        "application/vnd.oci.image.manifest.v1+json",
        "application/vnd.docker.distribution.manifest.v2+json",
    ]
)

def get_manifest_digest(registry: str, repository: str, reference: str) -> str | None:
    url = f"https://{registry}/v2/{repository}/manifests/{reference}"
    headers = {"Accept": _ACCEPT_HEADERS}
    status, response_headers = head_request(url, headers)
    digest = _read_digest(response_headers)
    if digest:
        return digest
    if status not in {401, 403}:
        return None

    auth_header = get_header(response_headers, "www-authenticate")
    if not auth_header:
        return None

    scheme, params = parse_auth_header(auth_header)
    if scheme != "bearer":
        return None
    token = fetch_bearer_token(
        realm=params.get("realm", ""),
        service=params.get("service", ""),
        scope=params.get("scope") or f"repository:{repository}:pull",
    )
    if not token:
        return None

    auth_headers = {"Accept": _ACCEPT_HEADERS, "Authorization": f"Bearer {token}"}
    _, response_headers = head_request(url, auth_headers)
    return _read_digest(response_headers)


def _read_digest(headers: Mapping[str, str]) -> str | None:
    digest = get_header(headers, "docker-content-digest")
    if digest:
        return digest
    return None



from typing import Any

from aicage.config._schema_validation import load_schema, validate_schema_mapping
from aicage.config._yaml import expect_bool, expect_string
from aicage.config.base.models import BUILD_LOCAL_KEY
from aicage.config.errors import ConfigError

_BASE_SCHEMA_PATH: str = "validation/base.schema.json"
_BASE_CONTEXT: str = "base metadata"


def validate_base_mapping(mapping: dict[str, Any]) -> dict[str, Any]:
    schema = load_schema(_BASE_SCHEMA_PATH)
    return validate_schema_mapping(
        mapping,
        schema,
        _BASE_CONTEXT,
        normalizer=_apply_defaults,
        value_validator=_validate_value,
    )


def _apply_defaults(mapping: dict[str, Any]) -> dict[str, Any]:
    normalized = dict(mapping)
    normalized.setdefault(BUILD_LOCAL_KEY, False)
    return normalized


def _validate_value(value: Any, schema_entry: dict[str, Any], context: str) -> None:
    schema_type = schema_entry.get("type")
    if schema_type == "string":
        expect_string(value, context)
        return
    if schema_type == "boolean":
        expect_bool(value, context)
        return
    raise ConfigError(f"{context} has unsupported schema type '{schema_type}'.")



import os
from dataclasses import dataclass
from pathlib import Path

from aicage.config.agent.models import AgentMetadata


@dataclass
class AgentConfig:
    agent_path: str
    agent_config_host: Path


def resolve_agent_config(agent_metadata: AgentMetadata) -> AgentConfig:
    agent_path = agent_metadata.agent_path
    agent_config_host = Path(os.path.expanduser(agent_path)).resolve()
    agent_config_host.mkdir(parents=True, exist_ok=True)
    return AgentConfig(agent_path=agent_path, agent_config_host=agent_config_host)



from dataclasses import dataclass

from aicage._logging import get_logger
from aicage.config.agent.models import AgentMetadata
from aicage.config.base.filter import filter_bases
from aicage.config.context import ConfigContext
from aicage.constants import DEFAULT_IMAGE_BASE
from aicage.runtime._errors import RuntimeExecutionError

from ._tty import ensure_tty_for_prompt


@dataclass(frozen=True)
class BaseSelectionRequest:
    agent: str
    context: ConfigContext
    agent_metadata: AgentMetadata


@dataclass(frozen=True)
class BaseOption:
    base: str
    description: str


def prompt_for_base(request: BaseSelectionRequest) -> str:
    ensure_tty_for_prompt()
    logger = get_logger()
    title = f"Select base image for '{request.agent}' (runtime to use inside the container):"
    bases = base_options(request.context, request.agent_metadata)

    if bases:
        print(title)
        for idx, option in enumerate(bases, start=1):
            suffix = " (default)" if option.base == DEFAULT_IMAGE_BASE else ""
            print(f"  {idx}) {option.base}: {option.description}{suffix}")
        prompt = f"Enter number or name [{DEFAULT_IMAGE_BASE}]: "
    else:
        prompt = f"{title} [{DEFAULT_IMAGE_BASE}]: "

    response = input(prompt).strip()
    if not response:
        choice = DEFAULT_IMAGE_BASE
    elif response.isdigit() and bases:
        idx = int(response)
        if idx < 1 or idx > len(bases):
            raise RuntimeExecutionError(
                f"Invalid choice '{response}'. Pick a number between 1 and {len(bases)}."
            )
        choice = bases[idx - 1].base
    else:
        choice = response

    if bases and choice not in available_bases(bases):
        options = ", ".join(available_bases(bases))
        raise RuntimeExecutionError(f"Invalid base '{choice}'. Valid options: {options}")
    logger.info("Selected base '%s' for agent '%s'", choice, request.agent)
    return choice


def base_options(context: ConfigContext, agent_metadata: AgentMetadata) -> list[BaseOption]:
    bases = filter_bases(context, agent_metadata)
    return [
        BaseOption(
            base=base,
            description=context.bases[base].base_image_description,
        )
        for base in sorted(bases)
    ]


def available_bases(options: list[BaseOption]) -> list[str]:
    return [option.base for option in options]



from pathlib import Path
from typing import Any

from aicage.config._yaml import expect_bool, expect_string, maybe_str_list
from aicage.config.agent._validation import validate_agent_mapping
from aicage.config.agent.models import (
    AGENT_FULL_NAME_KEY,
    AGENT_HOMEPAGE_KEY,
    AGENT_PATH_KEY,
    BASE_DISTRO_EXCLUDE_KEY,
    BASE_EXCLUDE_KEY,
    BUILD_LOCAL_KEY,
    AgentMetadata,
)
from aicage.config.base.models import BaseMetadata
from aicage.config.image_refs import local_image_ref
from aicage.constants import IMAGE_REGISTRY, IMAGE_REPOSITORY, LOCAL_IMAGE_REPOSITORY


def build_agent_metadata(
    agent_name: str,
    agent_mapping: dict[str, Any],
    bases: dict[str, BaseMetadata],
    definition_dir: Path,
) -> AgentMetadata:
    normalized_mapping = validate_agent_mapping(agent_mapping)
    base_exclude = maybe_str_list(normalized_mapping.get(BASE_EXCLUDE_KEY), BASE_EXCLUDE_KEY) or []
    base_distro_exclude = (
        maybe_str_list(normalized_mapping.get(BASE_DISTRO_EXCLUDE_KEY), BASE_DISTRO_EXCLUDE_KEY)
        or []
    )
    build_local = expect_bool(normalized_mapping.get(BUILD_LOCAL_KEY), BUILD_LOCAL_KEY)
    valid_bases = _build_valid_bases(
        agent_name=agent_name,
        bases=bases,
        base_exclude=base_exclude,
        base_distro_exclude=base_distro_exclude,
        build_local=build_local,
    )
    return AgentMetadata(
        agent_path=expect_string(normalized_mapping.get(AGENT_PATH_KEY), AGENT_PATH_KEY),
        agent_full_name=expect_string(normalized_mapping.get(AGENT_FULL_NAME_KEY), AGENT_FULL_NAME_KEY),
        agent_homepage=expect_string(normalized_mapping.get(AGENT_HOMEPAGE_KEY), AGENT_HOMEPAGE_KEY),
        build_local=build_local,
        valid_bases=valid_bases,
        base_exclude=base_exclude,
        base_distro_exclude=base_distro_exclude,
        local_definition_dir=definition_dir,
    )


def _build_valid_bases(
    agent_name: str,
    bases: dict[str, BaseMetadata],
    base_exclude: list[str],
    base_distro_exclude: list[str],
    build_local: bool,
) -> dict[str, str]:
    valid_bases: dict[str, str] = {}
    base_exclude_set = _normalize_exclude(base_exclude)
    base_distro_exclude_set = _normalize_exclude(base_distro_exclude)
    repository = _image_repository(build_local)
    for base_name in sorted(bases):
        base_metadata = bases[base_name]
        if _is_base_excluded(
            base_name,
            base_metadata.base_image_distro,
            base_exclude_set,
            base_distro_exclude_set,
        ):
            continue
        valid_bases[base_name] = local_image_ref(repository, agent_name, base_name)
    return valid_bases


def _image_repository(build_local: bool) -> str:
    if build_local:
        return LOCAL_IMAGE_REPOSITORY
    return f"{IMAGE_REGISTRY}/{IMAGE_REPOSITORY}"


def _is_base_excluded(
    base_name: str,
    base_distro: str,
    base_exclude: set[str],
    base_distro_exclude: set[str],
) -> bool:
    base_name_lc = base_name.lower()
    if base_name_lc in base_exclude:
        return True
    if base_distro.lower() in base_distro_exclude:
        return True
    return False


def _normalize_exclude(values: list[str]) -> set[str]:
    if not values:
        return set()
    return {value.lower() for value in values}



from dataclasses import dataclass


@dataclass
class ParsedArgs:
    dry_run: bool
    docker_args: str
    agent: str
    agent_args: list[str]
    entrypoint: str | None
    docker_socket: bool
    config_action: str | None



from pathlib import Path

from aicage.config.agent.models import AgentMetadata
from aicage.config.runtime_config import RunConfig
from aicage.docker.build import run_build
from aicage.paths import CUSTOM_BASES_DIR
from aicage.registry._time import now_iso

from ..agent_version.checker import AgentVersionChecker
from ._custom_base import ensure_custom_base_image
from ._digest import refresh_base_digest
from ._logs import build_log_path
from ._plan import should_build
from ._refs import base_repository, get_base_image_ref
from ._store import BuildRecord, BuildStore


def ensure_local_image(run_config: RunConfig) -> None:
    agent_metadata = run_config.context.agents[run_config.agent]
    definition_dir = agent_metadata.local_definition_dir

    base_metadata = run_config.context.bases[run_config.selection.base]
    custom_base = base_metadata.local_definition_dir.is_relative_to(CUSTOM_BASES_DIR)
    base_image = get_base_image_ref(run_config)
    image_ref = run_config.selection.base_image_ref
    if custom_base:
        ensure_custom_base_image(
            run_config.selection.base,
            base_metadata,
            base_metadata.local_definition_dir,
        )
    else:
        base_repo = base_repository(run_config)
        refresh_base_digest(
            base_image_ref=base_image,
            base_repository=base_repo,
        )

    store = BuildStore()
    record = store.load(run_config.agent, run_config.selection.base)

    agent_version = _get_agent_version(run_config, agent_metadata, definition_dir)
    needs_build = should_build(
        run_config=run_config,
        record=record,
        agent_version=agent_version,
    )
    if not needs_build:
        return

    log_path = build_log_path(run_config.agent, run_config.selection.base)
    run_build(
        run_config=run_config,
        base_image_ref=base_image,
        image_ref=image_ref,
        log_path=log_path,
    )

    store.save(
        BuildRecord(
            agent=run_config.agent,
            base=run_config.selection.base,
            agent_version=agent_version,
            base_image=base_image,
            image_ref=image_ref,
            built_at=now_iso(),
        )
    )


def _get_agent_version(
    run_config: RunConfig,
    agent_metadata: AgentMetadata,
    definition_dir: Path,
) -> str:
    checker = AgentVersionChecker()
    return checker.get_version(
        run_config.agent,
        agent_metadata,
        definition_dir,
    )



import json
import re
import urllib.error
import urllib.parse
import urllib.request

from ._timeouts import REGISTRY_REQUEST_TIMEOUT_SECONDS

_AUTH_HEADER_SPLIT_PARTS: int = 2


def parse_auth_header(value: str) -> tuple[str, dict[str, str]]:
    parts = value.split(" ", 1)
    if not parts:
        return "", {}
    scheme = parts[0].strip().lower()
    params: dict[str, str] = {}
    if len(parts) == _AUTH_HEADER_SPLIT_PARTS:
        params = dict(re.findall(r'(\w+)="([^"]+)"', parts[1]))
    return scheme, params


def fetch_bearer_token(realm: str, service: str, scope: str) -> str | None:
    if not realm:
        return None
    query = {"service": service, "scope": scope} if service else {"scope": scope}
    url = f"{realm}?{urllib.parse.urlencode(query)}"
    request = urllib.request.Request(url, headers={"Accept": "application/json"})
    try:
        with urllib.request.urlopen(request, timeout=REGISTRY_REQUEST_TIMEOUT_SECONDS) as response:
            payload = response.read().decode("utf-8")
    except urllib.error.URLError:
        return None
    try:
        data = json.loads(payload)
    except json.JSONDecodeError:
        return None
    token = data.get("token") or data.get("access_token")
    if not isinstance(token, str) or not token:
        return None
    return token



from typing import Any

from aicage.config._schema_validation import load_schema, validate_schema_mapping
from aicage.config._yaml import expect_string
from aicage.config.errors import ConfigError

_EXTENSION_SCHEMA_PATH = "validation/extension.schema.json"
_EXTENSION_CONTEXT = "extension metadata"


def validate_extension_mapping(mapping: dict[str, Any]) -> dict[str, Any]:
    context = _EXTENSION_CONTEXT
    schema = load_schema(_EXTENSION_SCHEMA_PATH)
    return validate_schema_mapping(mapping, schema, context, value_validator=_validate_value)


def _validate_value(value: Any, schema_entry: dict[str, Any], context: str) -> None:
    schema_type = schema_entry.get("type")
    if schema_type == "string":
        expect_string(value, context)
        return
    raise ConfigError(f"{context} has unsupported schema type '{schema_type}'.")



import hashlib
from dataclasses import dataclass
from pathlib import Path
from typing import Protocol

from aicage.config._yaml import expect_string
from aicage.config.errors import ConfigError
from aicage.config.extensions._validation import validate_extension_mapping
from aicage.config.yaml_loader import load_yaml
from aicage.paths import CUSTOM_EXTENSION_DEFINITION_FILES, CUSTOM_EXTENSIONS_DIR

_EXTENSION_NAME_KEY: str = "name"
_EXTENSION_DESCRIPTION_KEY: str = "description"
_SCRIPTS_DIRNAME: str = "scripts"
_DOCKERFILE_NAME: str = "Dockerfile"


class _HashWriter(Protocol):
    def update(self, data: bytes, /) -> None:
        ...

    def hexdigest(self) -> str:
        ...


@dataclass(frozen=True)
class ExtensionMetadata:
    extension_id: str
    name: str
    description: str
    directory: Path
    scripts_dir: Path
    dockerfile_path: Path | None


def load_extensions() -> dict[str, ExtensionMetadata]:
    extensions_dir = CUSTOM_EXTENSIONS_DIR
    if not extensions_dir.is_dir():
        return {}
    extensions: dict[str, ExtensionMetadata] = {}
    for entry in sorted(extensions_dir.iterdir()):
        if not entry.is_dir():
            continue
        extension_id = entry.name
        definition_path = _find_extension_definition(entry)
        mapping = validate_extension_mapping(load_yaml(definition_path))
        scripts_dir = entry / _SCRIPTS_DIRNAME
        if not scripts_dir.is_dir():
            raise ConfigError(f"Extension '{extension_id}' is missing scripts/ directory.")
        dockerfile_path = entry / _DOCKERFILE_NAME
        extensions[extension_id] = ExtensionMetadata(
            extension_id=extension_id,
            name=expect_string(mapping.get(_EXTENSION_NAME_KEY), _EXTENSION_NAME_KEY),
            description=expect_string(mapping.get(_EXTENSION_DESCRIPTION_KEY), _EXTENSION_DESCRIPTION_KEY),
            directory=entry,
            scripts_dir=scripts_dir,
            dockerfile_path=dockerfile_path if dockerfile_path.is_file() else None,
        )
    return extensions


def extension_hash(extension: ExtensionMetadata) -> str:
    digest = hashlib.sha256()
    definition_path = _find_extension_definition(extension.directory)
    _update_hash(digest, definition_path)
    if extension.dockerfile_path is not None:
        _update_hash(digest, extension.dockerfile_path)
    for script in sorted(extension.scripts_dir.glob("*.sh")):
        if script.is_file():
            _update_hash(digest, script)
    return digest.hexdigest()


def _find_extension_definition(extension_dir: Path) -> Path:
    for filename in CUSTOM_EXTENSION_DEFINITION_FILES:
        candidate = extension_dir / filename
        if candidate.is_file():
            return candidate
    expected = ", ".join(CUSTOM_EXTENSION_DEFINITION_FILES)
    raise ConfigError(f"Extension '{extension_dir.name}' is missing {expected}.")


def _update_hash(digest: _HashWriter, path: Path) -> None:
    digest.update(path.name.encode("utf-8"))
    digest.update(path.read_bytes())



from pathlib import Path

from aicage._logging import get_logger
from aicage.config.config_store import SettingsStore


def print_project_config() -> None:
    logger = get_logger()
    store = SettingsStore()
    project_path = Path.cwd().resolve()
    config_path = store.project_config_path(project_path)
    logger.info("Printing project config at %s", config_path)
    print("Project config path:")
    print(config_path)
    print()
    print("Project config content:")
    if config_path.exists():
        contents = config_path.read_text(encoding="utf-8").rstrip()
        if contents:
            print(contents)
        else:
            print("(empty)")
    else:
        print("(missing)")



import hashlib
from pathlib import Path
from typing import Any

import yaml

from aicage.paths import PROJECTS_DIR

from .project_config import ProjectConfig
from .yaml_loader import load_yaml


class SettingsStore:
    """
    Persists per-project configuration under ~/.aicage.
    """

    def __init__(self) -> None:
        self.projects_dir = PROJECTS_DIR
        self.projects_dir.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def _save_yaml(path: Path, data: dict[str, Any]) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w", encoding="utf-8") as handle:
            yaml.safe_dump(data, handle, sort_keys=True)

    def _project_path(self, project_realpath: Path) -> Path:
        digest = hashlib.sha256(str(project_realpath).encode("utf-8")).hexdigest()
        return self.projects_dir / f"{digest}.yaml"

    def load_project(self, project_realpath: Path) -> ProjectConfig:
        path = self._project_path(project_realpath)
        if not path.exists():
            data = {}
        else:
            data = load_yaml(path)
        return ProjectConfig.from_mapping(project_realpath, data)

    def save_project(self, project_realpath: Path, config: ProjectConfig) -> None:
        self._save_yaml(self._project_path(project_realpath), config.to_mapping())

    def project_config_path(self, project_realpath: Path) -> Path:
        """
        Returns the path to a project's config file under the base directory.
        """
        return self._project_path(project_realpath)



import sys

from aicage.runtime._errors import RuntimeExecutionError


def ensure_tty_for_prompt() -> None:
    if not sys.stdin.isatty():
        raise RuntimeExecutionError("Interactive input required but stdin is not a TTY.")



from aicage.config.runtime_config import RunConfig
from aicage.constants import IMAGE_BASE_REPOSITORY, IMAGE_REGISTRY
from aicage.paths import CUSTOM_BASES_DIR

from ._custom_base import custom_base_image_ref


def get_base_image_ref(run_config: RunConfig) -> str:
    base_metadata = run_config.context.bases[run_config.selection.base]
    if base_metadata.local_definition_dir.is_relative_to(CUSTOM_BASES_DIR):
        return custom_base_image_ref(run_config.selection.base)
    repository = base_repository(run_config)
    return f"{repository}:{run_config.selection.base}"


def base_repository(_run_config: RunConfig) -> str:
    return f"{IMAGE_REGISTRY}/{IMAGE_BASE_REPOSITORY}"



from pathlib import Path
from typing import Any

from aicage.config._schema_validation import load_schema, validate_schema_mapping
from aicage.config._yaml import expect_bool, expect_string
from aicage.config.agent.models import BUILD_LOCAL_KEY
from aicage.config.errors import ConfigError

_AGENT_SCHEMA_PATH = "validation/agent.schema.json"
_AGENT_CONTEXT = "agent metadata"


def validate_agent_mapping(mapping: dict[str, Any]) -> dict[str, Any]:
    schema = load_schema(_AGENT_SCHEMA_PATH)
    return validate_schema_mapping(
        mapping,
        schema,
        _AGENT_CONTEXT,
        normalizer=_apply_defaults,
        value_validator=_validate_value,
    )


def ensure_required_files(agent_name: str, agent_dir: Path) -> None:
    missing = [name for name in ("install.sh", "version.sh") if not (agent_dir / name).is_file()]
    if missing:
        raise ConfigError(f"Agent '{agent_name}' is missing {', '.join(missing)}.")


def _apply_defaults(mapping: dict[str, Any]) -> dict[str, Any]:
    normalized = dict(mapping)
    normalized.setdefault(BUILD_LOCAL_KEY, True)
    return normalized


def _validate_value(value: Any, schema_entry: dict[str, Any], context: str) -> None:
    schema_type = schema_entry.get("type")
    if schema_type == "string":
        expect_string(value, context)
        return
    if schema_type == "boolean":
        expect_bool(value, context)
        return
    if schema_type == "array":
        _expect_str_list(value, context, schema_entry)
        return
    raise ConfigError(f"{context} has unsupported schema type '{schema_type}'.")


def _expect_str_list(value: Any, context: str, schema_entry: dict[str, Any]) -> None:
    if not isinstance(value, list):
        raise ConfigError(f"{context} must be a list.")
    item_schema = schema_entry.get("items", {})
    item_type = item_schema.get("type")
    if item_type != "string":
        raise ConfigError(f"{context} items must be strings.")
    for item in value:
        expect_string(item, context)
